@@ -29,6 +29,38 @@ use crate::RunCtx;
 
 type Result<T> = result::Result<T, Error>;
 
+/// Wraps a writer and buffers its output.
+///
+/// This implementation is similar to `io::BufWriter`, but:
+/// - The inner writer is only written to if `flush` is called.
+/// - The buffer may grow infinitely large.
+struct BufWriter<'w, W: io::Write + ?Sized> {
+  buf: Vec<u8>,
+  inner: &'w mut W,
+}
+
+impl<'w, W: io::Write + ?Sized> BufWriter<'w, W> {
+  pub fn new(inner: &'w mut W) -> Self {
+    BufWriter {
+      buf: Vec::with_capacity(128),
+      inner,
+    }
+  }
+}
+
+impl<'w, W: io::Write + ?Sized> io::Write for BufWriter<'w, W> {
+  fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
+    self.buf.extend_from_slice(buf);
+    Ok(buf.len())
+  }
+
+  fn flush(&mut self) -> io::Result<()> {
+    self.inner.write_all(&self.buf)?;
+    self.buf.clear();
+    self.inner.flush()
+  }
+}
+
 trait Stdio {
   fn stdio(&mut self) -> (&mut dyn io::Write, &mut dyn io::Write);
 }
@@ -39,6 +71,15 @@ impl<'io> Stdio for RunCtx<'io> {
   }
 }
 
+impl<W> Stdio for (&mut W, &mut W)
+where
+  W: io::Write,
+{
+  fn stdio(&mut self) -> (&mut dyn io::Write, &mut dyn io::Write) {
+    (self.0, self.1)
+  }
+}
+
 /// A command execution context that captures additional data pertaining
 /// the command execution.
 pub struct ExecCtx<'io> {
@@ -49,6 +90,8 @@ pub struct ExecCtx<'io> {
   pub user_pin: Option<ffi::OsString>,
   pub new_admin_pin: Option<ffi::OsString>,
   pub new_user_pin: Option<ffi::OsString>,
+  pub password: Option<ffi::OsString>,
+  pub no_cache: bool,
   pub verbosity: u64,
 }
 
@@ -59,51 +102,50 @@ impl<'io> Stdio for ExecCtx<'io> {
 }
 
 /// The available Nitrokey models.
+#[allow(unused_doc_comments)]
 Enum! {DeviceModel, [
   Pro => "pro",
-  Storage => "storage"
+  Storage => "storage",
 ]}
 
+impl DeviceModel {
+  pub fn as_user_facing_str(&self) -> &str {
+    match self {
+      DeviceModel::Pro => "Pro",
+      DeviceModel::Storage => "Storage",
+    }
+  }
+}
+
+impl From<DeviceModel> for nitrokey::Model {
+  fn from(model: DeviceModel) -> nitrokey::Model {
+    match model {
+      DeviceModel::Pro => nitrokey::Model::Pro,
+      DeviceModel::Storage => nitrokey::Model::Storage,
+    }
+  }
+}
+
 /// A top-level command for nitrocli.
+#[allow(unused_doc_comments)]
 Enum! {Command, [
-  Config => "config",
-  Lock => "lock",
-  Otp => "otp",
-  Pin => "pin",
-  Pws => "pws",
-  Status => "status",
-  Storage => "storage"
+  Config => ("config", config),
+  Encrypted => ("encrypted", encrypted),
+  Hidden => ("hidden", hidden),
+  Lock => ("lock", lock),
+  Otp => ("otp", otp),
+  Pin => ("pin", pin),
+  Pws => ("pws", pws),
+  Reset => ("reset", reset),
+  Status => ("status", status),
+  Unencrypted => ("unencrypted", unencrypted),
 ]}
 
-impl Command {
-  /// Execute this command with the given arguments.
-  pub fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
-    match self {
-      Command::Config => config(ctx, args),
-      Command::Lock => lock(ctx, args),
-      Command::Otp => otp(ctx, args),
-      Command::Pin => pin(ctx, args),
-      Command::Pws => pws(ctx, args),
-      Command::Status => status(ctx, args),
-      Command::Storage => storage(ctx, args),
-    }
-  }
-}
-
 Enum! {ConfigCommand, [
-  Get => "get",
-  Set => "set"
+  Get => ("get", config_get),
+  Set => ("set", config_set),
 ]}
 
-impl ConfigCommand {
-  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
-    match self {
-      ConfigCommand::Get => config_get(ctx, args),
-      ConfigCommand::Set => config_set(ctx, args),
-    }
-  }
-}
-
 #[derive(Clone, Copy, Debug)]
 pub enum ConfigOption<T> {
   Enable(T),
@@ -140,31 +182,20 @@ impl<T> ConfigOption<T> {
 }
 
 Enum! {OtpCommand, [
-  Clear => "clear",
-  Get => "get",
-  Set => "set",
-  Status => "status"
+  Clear => ("clear", otp_clear),
+  Get => ("get", otp_get),
+  Set => ("set", otp_set),
+  Status => ("status", otp_status),
 ]}
 
-impl OtpCommand {
-  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
-    match self {
-      OtpCommand::Clear => otp_clear(ctx, args),
-      OtpCommand::Get => otp_get(ctx, args),
-      OtpCommand::Set => otp_set(ctx, args),
-      OtpCommand::Status => otp_status(ctx, args),
-    }
-  }
-}
-
 Enum! {OtpAlgorithm, [
   Hotp => "hotp",
-  Totp => "totp"
+  Totp => "totp",
 ]}
 
 Enum! {OtpMode, [
   SixDigits => "6",
-  EightDigits => "8"
+  EightDigits => "8",
 ]}
 
 impl From<OtpMode> for nitrokey::OtpMode {
@@ -179,153 +210,261 @@ impl From<OtpMode> for nitrokey::OtpMode {
 Enum! {OtpSecretFormat, [
   Ascii => "ascii",
   Base32 => "base32",
-  Hex => "hex"
+  Hex => "hex",
 ]}
 
 Enum! {PinCommand, [
-  Clear => "clear",
-  Set => "set",
-  Unblock => "unblock"
+  Clear => ("clear", pin_clear),
+  Set => ("set", pin_set),
+  Unblock => ("unblock", pin_unblock),
 ]}
 
-impl PinCommand {
-  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
-    match self {
-      PinCommand::Clear => pin_clear(ctx, args),
-      PinCommand::Set => pin_set(ctx, args),
-      PinCommand::Unblock => pin_unblock(ctx, args),
-    }
-  }
-}
-
 Enum! {PwsCommand, [
-  Clear => "clear",
-  Get => "get",
-  Set => "set",
-  Status => "status"
+  Clear => ("clear", pws_clear),
+  Get => ("get", pws_get),
+  Set => ("set", pws_set),
+  Status => ("status", pws_status),
 ]}
 
-impl PwsCommand {
-  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
-    match self {
-      PwsCommand::Clear => pws_clear(ctx, args),
-      PwsCommand::Get => pws_get(ctx, args),
-      PwsCommand::Set => pws_set(ctx, args),
-      PwsCommand::Status => pws_status(ctx, args),
-    }
-  }
-}
-
 fn parse(
   ctx: &mut impl Stdio,
-  parser: &argparse::ArgumentParser<'_>,
+  parser: argparse::ArgumentParser<'_>,
   args: Vec<String>,
 ) -> Result<()> {
   let (stdout, stderr) = ctx.stdio();
-  if let Err(err) = parser.parse(args, stdout, stderr) {
-    Err(Error::ArgparseError(err))
-  } else {
-    Ok(())
-  }
+  let result = parser
+    .parse(args, stdout, stderr)
+    .map_err(Error::ArgparseError);
+  drop(parser);
+  result
 }
 
-/// Inquire the status of the nitrokey.
+/// Inquire the status of the Nitrokey.
 fn status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the status of the connected Nitrokey device");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::status(ctx)
 }
 
-Enum! {StorageCommand, [
-  Close => "close",
-  Open => "open",
-  Status => "status"
-]}
+/// Perform a factory reset.
+fn reset(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Performs a factory reset");
+  parse(ctx, parser, args)?;
 
-impl StorageCommand {
-  fn execute(self, ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
-    match self {
-      StorageCommand::Close => storage_close(ctx, args),
-      StorageCommand::Open => storage_open(ctx, args),
-      StorageCommand::Status => storage_status(ctx, args),
-    }
-  }
+  commands::reset(ctx)
 }
 
-/// Execute a storage subcommand.
-fn storage(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
-  let mut subcommand = StorageCommand::Open;
+Enum! {UnencryptedCommand, [
+  Set => ("set", unencrypted_set),
+]}
+
+Enum! {UnencryptedVolumeMode, [
+  ReadWrite => "read-write",
+  ReadOnly => "read-only",
+]}
+
+/// Execute an unencrypted subcommand.
+fn unencrypted(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut subcommand = UnencryptedCommand::Set;
+  let help = cmd_help!(subcommand);
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
-  parser.set_description("Interacts with the device's storage");
-  let _ = parser.refer(&mut subcommand).required().add_argument(
-    "subcommand",
-    argparse::Store,
-    "The subcommand to execute (open|close)",
-  );
+  parser.set_description("Interacts with the device's unencrypted volume");
+  let _ =
+    parser
+      .refer(&mut subcommand)
+      .required()
+      .add_argument("subcommand", argparse::Store, &help);
   let _ = parser.refer(&mut subargs).add_argument(
     "arguments",
     argparse::List,
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
-  subargs.insert(0, format!("nitrocli {} {}", Command::Storage, subcommand));
+  subargs.insert(
+    0,
+    format!(
+      "{} {} {}",
+      crate::NITROCLI,
+      Command::Unencrypted,
+      subcommand,
+    ),
+  );
   subcommand.execute(ctx, subargs)
 }
 
-/// Open the encrypted volume on the nitrokey.
-fn storage_open(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+/// Change the configuration of the unencrypted volume.
+fn unencrypted_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut mode = UnencryptedVolumeMode::ReadWrite;
+  let help = format!("The mode to change to ({})", fmt_enum!(mode));
+  let mut parser = argparse::ArgumentParser::new();
+  parser
+    .set_description("Changes the configuration of the unencrypted volume on a Nitrokey Storage");
+  let _ = parser
+    .refer(&mut mode)
+    .required()
+    .add_argument("type", argparse::Store, &help);
+  parse(ctx, parser, args)?;
+
+  commands::unencrypted_set(ctx, mode)
+}
+
+Enum! {EncryptedCommand, [
+  Close => ("close", encrypted_close),
+  Open => ("open", encrypted_open),
+]}
+
+/// Execute an encrypted subcommand.
+fn encrypted(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut subcommand = EncryptedCommand::Open;
+  let help = cmd_help!(subcommand);
+  let mut subargs = vec![];
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Interacts with the device's encrypted volume");
+  let _ =
+    parser
+      .refer(&mut subcommand)
+      .required()
+      .add_argument("subcommand", argparse::Store, &help);
+  let _ = parser.refer(&mut subargs).add_argument(
+    "arguments",
+    argparse::List,
+    "The arguments for the subcommand",
+  );
+  parser.stop_on_first_argument(true);
+  parse(ctx, parser, args)?;
+
+  subargs.insert(
+    0,
+    format!("{} {} {}", crate::NITROCLI, Command::Encrypted, subcommand),
+  );
+  subcommand.execute(ctx, subargs)
+}
+
+/// Open the encrypted volume on the Nitrokey.
+fn encrypted_open(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Opens the encrypted volume on a Nitrokey Storage");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
-  commands::storage_open(ctx)
+  commands::encrypted_open(ctx)
 }
 
 /// Close the previously opened encrypted volume.
-fn storage_close(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+fn encrypted_close(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Closes the encrypted volume on a Nitrokey Storage");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
-  commands::storage_close(ctx)
+  commands::encrypted_close(ctx)
 }
 
-/// Print the status of the nitrokey's storage.
-fn storage_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
-  let mut parser = argparse::ArgumentParser::new();
-  parser.set_description("Prints the status of the Nitrokey's storage");
-  parse(ctx, &parser, args)?;
+Enum! {HiddenCommand, [
+  Close => ("close", hidden_close),
+  Create => ("create", hidden_create),
+  Open => ("open", hidden_open),
+]}
 
-  commands::storage_status(ctx)
+/// Execute a hidden subcommand.
+fn hidden(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut subcommand = HiddenCommand::Open;
+  let help = cmd_help!(subcommand);
+  let mut subargs = vec![];
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Interacts with the device's hidden volume");
+  let _ =
+    parser
+      .refer(&mut subcommand)
+      .required()
+      .add_argument("subcommand", argparse::Store, &help);
+  let _ = parser.refer(&mut subargs).add_argument(
+    "arguments",
+    argparse::List,
+    "The arguments for the subcommand",
+  );
+  parser.stop_on_first_argument(true);
+  parse(ctx, parser, args)?;
+
+  subargs.insert(
+    0,
+    format!("{} {} {}", crate::NITROCLI, Command::Hidden, subcommand),
+  );
+  subcommand.execute(ctx, subargs)
+}
+
+fn hidden_create(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut slot: u8 = 0;
+  let mut start: u8 = 0;
+  let mut end: u8 = 0;
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Creates a hidden volume on a Nitrokey Storage");
+  let _ = parser.refer(&mut slot).required().add_argument(
+    "slot",
+    argparse::Store,
+    "The hidden volume slot to use",
+  );
+  let _ = parser.refer(&mut start).required().add_argument(
+    "start",
+    argparse::Store,
+    "The start location of the hidden volume as percentage of the \
+     encrypted volume's size (0-99)",
+  );
+  let _ = parser.refer(&mut end).required().add_argument(
+    "end",
+    argparse::Store,
+    "The end location of the hidden volume as percentage of the \
+     encrypted volume's size (1-100)",
+  );
+  parse(ctx, parser, args)?;
+
+  commands::hidden_create(ctx, slot, start, end)
+}
+
+fn hidden_open(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Opens a hidden volume on a Nitrokey Storage");
+  parse(ctx, parser, args)?;
+
+  commands::hidden_open(ctx)
+}
+
+fn hidden_close(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Closes the hidden volume on a Nitrokey Storage");
+  parse(ctx, parser, args)?;
+
+  commands::hidden_close(ctx)
 }
 
 /// Execute a config subcommand.
 fn config(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = ConfigCommand::Get;
+  let help = cmd_help!(subcommand);
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Reads or writes the device configuration");
-  let _ = parser.refer(&mut subcommand).required().add_argument(
-    "subcommand",
-    argparse::Store,
-    "The subcommand to execute (get|set)",
-  );
+  let _ =
+    parser
+      .refer(&mut subcommand)
+      .required()
+      .add_argument("subcommand", argparse::Store, &help);
   let _ = parser.refer(&mut subargs).add_argument(
     "arguments",
     argparse::List,
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
-  subargs.insert(0, format!("nitrocli {} {}", Command::Config, subcommand));
+  subargs.insert(
+    0,
+    format!("{} {} {}", crate::NITROCLI, Command::Config, subcommand),
+  );
   subcommand.execute(ctx, subargs)
 }
 
@@ -333,7 +472,7 @@ fn config(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn config_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Prints the Nitrokey configuration");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::config_get(ctx)
 }
@@ -390,8 +529,7 @@ fn config_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Allow one-time password generation without PIN",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   let numlock = ConfigOption::try_from(no_numlock, numlock, "numlock")?;
   let capslock = ConfigOption::try_from(no_capslock, capslock, "capslock")?;
@@ -410,7 +548,7 @@ fn config_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn lock(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Locks the connected Nitrokey device");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::lock(ctx)
 }
@@ -418,24 +556,27 @@ fn lock(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 /// Execute an OTP subcommand.
 fn otp(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = OtpCommand::Get;
+  let help = cmd_help!(subcommand);
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Accesses one-time passwords");
-  let _ = parser.refer(&mut subcommand).required().add_argument(
-    "subcommand",
-    argparse::Store,
-    "The subcommand to execute (clear|get|set|status)",
-  );
+  let _ =
+    parser
+      .refer(&mut subcommand)
+      .required()
+      .add_argument("subcommand", argparse::Store, &help);
   let _ = parser.refer(&mut subargs).add_argument(
     "arguments",
     argparse::List,
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
-  subargs.insert(0, format!("nitrocli {} {}", Command::Otp, subcommand));
+  subargs.insert(
+    0,
+    format!("{} {} {}", crate::NITROCLI, Command::Otp, subcommand),
+  );
   subcommand.execute(ctx, subargs)
 }
 
@@ -443,6 +584,11 @@ fn otp(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn otp_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut algorithm = OtpAlgorithm::Totp;
+  let help = format!(
+    "The OTP algorithm to use ({}, default: {})",
+    fmt_enum!(algorithm),
+    algorithm
+  );
   let mut time: Option<u64> = None;
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Generates a one-time password");
@@ -451,18 +597,15 @@ fn otp_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
       .refer(&mut slot)
       .required()
       .add_argument("slot", argparse::Store, "The OTP slot to use");
-  let _ = parser.refer(&mut algorithm).add_option(
-    &["-a", "--algorithm"],
-    argparse::Store,
-    "The OTP algorithm to use (hotp|totp)",
-  );
+  let _ = parser
+    .refer(&mut algorithm)
+    .add_option(&["-a", "--algorithm"], argparse::Store, &help);
   let _ = parser.refer(&mut time).add_option(
     &["-t", "--time"],
     argparse::StoreOption,
     "The time to use for TOTP generation (Unix timestamp, default: system time)",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::otp_get(ctx, slot, algorithm, time)
 }
@@ -471,13 +614,22 @@ fn otp_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 pub fn otp_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut algorithm = OtpAlgorithm::Totp;
+  let algo_help = format!(
+    "The OTP algorithm to use ({}, default: {})",
+    fmt_enum!(algorithm),
+    algorithm
+  );
   let mut name = "".to_owned();
   let mut secret = "".to_owned();
   let mut digits = OtpMode::SixDigits;
   let mut counter: u64 = 0;
   let mut time_window: u16 = 30;
-  let mut ascii = false;
-  let mut secret_format: Option<OtpSecretFormat> = None;
+  let mut secret_format = OtpSecretFormat::Hex;
+  let fmt_help = format!(
+    "The format of the secret ({}, default: {})",
+    fmt_enum!(OtpSecretFormat::all_variants()),
+    secret_format,
+  );
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Configures a one-time password slot");
   let _ =
@@ -485,11 +637,10 @@ pub fn otp_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
       .refer(&mut slot)
       .required()
       .add_argument("slot", argparse::Store, "The OTP slot to use");
-  let _ = parser.refer(&mut algorithm).add_option(
-    &["-a", "--algorithm"],
-    argparse::Store,
-    "The OTP algorithm to use (hotp or totp, default: totp)",
-  );
+  let _ =
+    parser
+      .refer(&mut algorithm)
+      .add_option(&["-a", "--algorithm"], argparse::Store, &algo_help);
   let _ = parser.refer(&mut name).required().add_argument(
     "name",
     argparse::Store,
@@ -498,7 +649,7 @@ pub fn otp_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let _ = parser.refer(&mut secret).required().add_argument(
     "secret",
     argparse::Store,
-    "The secret to store on the slot as a hexadecimal string (unless --ascii is set)",
+    "The secret to store on the slot as a hexadecimal string (unless overwritten by --format)",
   );
   let _ = parser.refer(&mut digits).add_option(
     &["-d", "--digits"],
@@ -515,33 +666,11 @@ pub fn otp_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The time window for TOTP (default: 30)",
   );
-  let _ = parser.refer(&mut ascii).add_option(
-    &["--ascii"],
-    argparse::StoreTrue,
-    "Interpret the given secret as an ASCII string of the secret (deprecated, use --format instead)"
-  );
-  let _ = parser.refer(&mut secret_format).add_option(
-    &["-f", "--format"],
-    argparse::StoreOption,
-    "The format of the secret (ascii|base32|hex)",
-  );
-  parse(ctx, &parser, args)?;
-  drop(parser);
-
-  if ascii {
-    if secret_format.is_some() {
-      return Err(Error::Error(
-        "The --format and the --ascii option cannot be used at the same time".to_string(),
-      ));
-    }
-
-    println!(
-      ctx,
-      "Warning: The --ascii option is deprecated. Please use --format ascii instead."
-    )?;
-    secret_format = Some(OtpSecretFormat::Ascii);
-  }
-  let secret_format = secret_format.unwrap_or(OtpSecretFormat::Hex);
+  let _ =
+    parser
+      .refer(&mut secret_format)
+      .add_option(&["-f", "--format"], argparse::Store, &fmt_help);
+  parse(ctx, parser, args)?;
 
   let data = nitrokey::OtpSlotData {
     number: slot,
@@ -558,6 +687,11 @@ pub fn otp_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn otp_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut slot: u8 = 0;
   let mut algorithm = OtpAlgorithm::Totp;
+  let help = format!(
+    "The OTP algorithm to use ({}, default: {})",
+    fmt_enum!(algorithm),
+    algorithm
+  );
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Clears a one-time password slot");
   let _ = parser.refer(&mut slot).required().add_argument(
@@ -565,13 +699,10 @@ fn otp_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The OTP slot to clear",
   );
-  let _ = parser.refer(&mut algorithm).add_option(
-    &["-a", "--algorithm"],
-    argparse::Store,
-    "The OTP algorithm to use (hotp|totp)",
-  );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  let _ = parser
+    .refer(&mut algorithm)
+    .add_option(&["-a", "--algorithm"], argparse::Store, &help);
+  parse(ctx, parser, args)?;
 
   commands::otp_clear(ctx, slot, algorithm)
 }
@@ -586,8 +717,7 @@ fn otp_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Show slots that are not programmed",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::otp_status(ctx, all)
 }
@@ -595,24 +725,27 @@ fn otp_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 /// Execute a PIN subcommand.
 fn pin(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = PinCommand::Clear;
+  let help = cmd_help!(subcommand);
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Manages the Nitrokey PINs");
-  let _ = parser.refer(&mut subcommand).required().add_argument(
-    "subcommand",
-    argparse::Store,
-    "The subcommand to execute (clear|set|unblock)",
-  );
+  let _ =
+    parser
+      .refer(&mut subcommand)
+      .required()
+      .add_argument("subcommand", argparse::Store, &help);
   let _ = parser.refer(&mut subargs).add_argument(
     "arguments",
     argparse::List,
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
-  subargs.insert(0, format!("nitrocli {} {}", Command::Pin, subcommand));
+  subargs.insert(
+    0,
+    format!("{} {} {}", crate::NITROCLI, Command::Pin, subcommand),
+  );
   subcommand.execute(ctx, subargs)
 }
 
@@ -620,23 +753,22 @@ fn pin(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn pin_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Clears the cached PINs");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
-  commands::pin_clear()
+  commands::pin_clear(ctx)
 }
 
 /// Change a PIN.
 fn pin_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut pintype = pinentry::PinType::User;
+  let help = format!("The PIN type to change ({})", fmt_enum!(pintype));
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Changes a PIN");
-  let _ = parser.refer(&mut pintype).required().add_argument(
-    "type",
-    argparse::Store,
-    "The PIN type to change (admin|user)",
-  );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  let _ = parser
+    .refer(&mut pintype)
+    .required()
+    .add_argument("type", argparse::Store, &help);
+  parse(ctx, parser, args)?;
 
   commands::pin_set(ctx, pintype)
 }
@@ -645,7 +777,7 @@ fn pin_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn pin_unblock(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Unblocks and resets the user PIN");
-  parse(ctx, &parser, args)?;
+  parse(ctx, parser, args)?;
 
   commands::pin_unblock(ctx)
 }
@@ -654,23 +786,26 @@ fn pin_unblock(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
 fn pws(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let mut subcommand = PwsCommand::Get;
   let mut subargs = vec![];
+  let help = cmd_help!(subcommand);
   let mut parser = argparse::ArgumentParser::new();
   parser.set_description("Accesses the password safe");
-  let _ = parser.refer(&mut subcommand).required().add_argument(
-    "subcommand",
-    argparse::Store,
-    "The subcommand to execute (clear|get|set|status)",
-  );
+  let _ =
+    parser
+      .refer(&mut subcommand)
+      .required()
+      .add_argument("subcommand", argparse::Store, &help);
   let _ = parser.refer(&mut subargs).add_argument(
     "arguments",
     argparse::List,
     "The arguments for the subcommand",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
-  subargs.insert(0, format!("nitrocli {} {}", Command::Pws, subcommand));
+  subargs.insert(
+    0,
+    format!("{} {} {}", crate::NITROCLI, Command::Pws, subcommand),
+  );
   subcommand.execute(ctx, subargs)
 }
 
@@ -708,8 +843,7 @@ fn pws_get(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Print the stored data without description",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pws_get(ctx, slot, name, login, password, quiet)
 }
@@ -725,7 +859,7 @@ fn pws_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
   let _ = parser.refer(&mut slot).required().add_argument(
     "slot",
     argparse::Store,
-    "The PWS slot to read",
+    "The PWS slot to write",
   );
   let _ = parser.refer(&mut name).required().add_argument(
     "name",
@@ -742,8 +876,7 @@ fn pws_set(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The password to store on the slot",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pws_set(ctx, slot, &name, &login, &password)
 }
@@ -758,8 +891,7 @@ fn pws_clear(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::Store,
     "The PWS slot to clear",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pws_clear(ctx, slot)
 }
@@ -774,65 +906,79 @@ fn pws_status(ctx: &mut ExecCtx<'_>, args: Vec<String>) -> Result<()> {
     argparse::StoreTrue,
     "Show slots that are not programmed",
   );
-  parse(ctx, &parser, args)?;
-  drop(parser);
+  parse(ctx, parser, args)?;
 
   commands::pws_status(ctx, all)
 }
 
-/// Parse the command-line arguments and return the selected command and
-/// the remaining arguments for the command.
-fn parse_arguments<'io, 'ctx: 'io>(
-  ctx: &'ctx mut RunCtx<'_>,
-  args: Vec<String>,
-) -> Result<(Command, ExecCtx<'io>, Vec<String>)> {
+/// Parse the command-line arguments and execute the selected command.
+pub(crate) fn handle_arguments(ctx: &mut RunCtx<'_>, args: Vec<String>) -> Result<()> {
+  use std::io::Write;
+
+  let mut version = false;
   let mut model: Option<DeviceModel> = None;
+  let model_help = format!(
+    "Select the device model to connect to ({})",
+    fmt_enum!(DeviceModel::all_variants())
+  );
   let mut verbosity = 0;
   let mut command = Command::Status;
+  let cmd_help = cmd_help!(command);
   let mut subargs = vec![];
   let mut parser = argparse::ArgumentParser::new();
+  let _ = parser.refer(&mut version).add_option(
+    &["-V", "--version"],
+    argparse::StoreTrue,
+    "Print version information and exit",
+  );
   let _ = parser.refer(&mut verbosity).add_option(
     &["-v", "--verbose"],
     argparse::IncrBy::<u64>(1),
     "Increase the log level (can be supplied multiple times)",
   );
-  let _ = parser.refer(&mut model).add_option(
-    &["-m", "--model"],
-    argparse::StoreOption,
-    "Select the device model to connect to (pro|storage)",
-  );
+  let _ =
+    parser
+      .refer(&mut model)
+      .add_option(&["-m", "--model"], argparse::StoreOption, &model_help);
   parser.set_description("Provides access to a Nitrokey device");
-  let _ = parser.refer(&mut command).required().add_argument(
-    "command",
-    argparse::Store,
-    "The command to execute (config|lock|otp|pin|pws|status|storage)",
-  );
+  let _ = parser
+    .refer(&mut command)
+    .required()
+    .add_argument("command", argparse::Store, &cmd_help);
   let _ = parser.refer(&mut subargs).add_argument(
     "arguments",
     argparse::List,
     "The arguments for the command",
   );
   parser.stop_on_first_argument(true);
-  parse(ctx, &parser, args)?;
-  drop(parser);
 
-  subargs.insert(0, format!("nitrocli {}", command));
+  let mut stdout_buf = BufWriter::new(ctx.stdout);
+  let mut stderr_buf = BufWriter::new(ctx.stderr);
+  let mut stdio_buf = (&mut stdout_buf, &mut stderr_buf);
+  let result = parse(&mut stdio_buf, parser, args);
 
-  let ctx = ExecCtx {
-    model,
-    stdout: ctx.stdout,
-    stderr: ctx.stderr,
-    admin_pin: ctx.admin_pin.take(),
-    user_pin: ctx.user_pin.take(),
-    new_admin_pin: ctx.new_admin_pin.take(),
-    new_user_pin: ctx.new_user_pin.take(),
-    verbosity,
-  };
-  Ok((command, ctx, subargs))
-}
+  if version {
+    println!(ctx, "{} {}", crate::NITROCLI, env!("CARGO_PKG_VERSION"))?;
+    Ok(())
+  } else {
+    stdout_buf.flush()?;
+    stderr_buf.flush()?;
 
-/// Parse the command-line arguments and execute the selected command.
-pub(crate) fn handle_arguments(ctx: &mut RunCtx<'_>, args: Vec<String>) -> Result<()> {
-  let (command, mut ctx, args) = parse_arguments(ctx, args)?;
-  command.execute(&mut ctx, args)
+    result?;
+    subargs.insert(0, format!("{} {}", crate::NITROCLI, command));
+
+    let mut ctx = ExecCtx {
+      model,
+      stdout: ctx.stdout,
+      stderr: ctx.stderr,
+      admin_pin: ctx.admin_pin.take(),
+      user_pin: ctx.user_pin.take(),
+      new_admin_pin: ctx.new_admin_pin.take(),
+      new_user_pin: ctx.new_user_pin.take(),
+      password: ctx.password.take(),
+      no_cache: ctx.no_cache,
+      verbosity,
+    };
+    command.execute(&mut ctx, subargs)
+  }
 }
